/*

                  c Ivo Hofacker

                  Vienna RNA package
*/
#include <config.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include "energy_par.h"
#include "fold_vars.h"
#include "utils.h"
#include "params.h"
/**
*** \file params.c
*** <P>
*** This file provides functions that return temperature scaled energy parameters and
*** Boltzmann weights packed in datastructures
*** </P>
***/

/*@unused@*/
static char rcsid[] UNUSED = "$Id: params.c,v 1.9 2008/07/04 14:29:14 ivo Exp $";
static int rtype[32] = {
                          0, 2, 1, 4, 3, 6, 5, 7,
                          0,18,17,20,19,22,21, 7,
                          0,10, 9,12,11,14,13, 7,
                          0,26,25,28,27,30,29, 7};
PRIVATE paramT p;
PRIVATE int id=-1;
/* variables for partition function */
PRIVATE pf_paramT pf;
PRIVATE int pf_id=-1;

#ifdef _OPENMP
#pragma omp threadprivate(id, pf_id)
#endif

PUBLIC paramT *scale_parameters(void){
  model_detailsT  md;
  set_model_details(&md);
  return get_scaled_parameters(temperature, md);
}

PUBLIC paramT *get_scaled_parameters( double temp,
                                      model_detailsT md){

  unsigned int i,j,k,l;
  double tempf;
  paramT *params;

  params  = (paramT *)space(sizeof(paramT));

  /* store the model details */
  params->model_details = md;
  params->temperature   = temp;
  tempf                 = ((params->temperature+K0)/Tmeasure);

  for(i = VRNA_GQUAD_MIN_STACK_SIZE; i <= VRNA_GQUAD_MAX_STACK_SIZE; i++)
    for(j = 3*VRNA_GQUAD_MIN_LINKER_LENGTH; j <= 3*VRNA_GQUAD_MAX_LINKER_LENGTH; j++){
      double GQuadAlpha_T = (double)GQuadAlphadH - (double)(GQuadAlphadH - GQuadAlpha37) * tempf;
      double GQuadBeta_T = (double)GQuadBetadH - (double)(GQuadBetadH - GQuadBeta37) * tempf;
      params->gquad[i][j] = (int)GQuadAlpha_T*(i-1) + (int)(((double)GQuadBeta_T)*log(j - 2));
    }

  for (i=0; i<31; i++)
    params->hairpin[i]  = hairpindH[i] - (hairpindH[i] - hairpin37[i])*tempf;
  for (i=0; i<=MIN2(30,MAXLOOP); i++) {
    params->bulge[i]          = bulgedH[i] - (bulgedH[i] - bulge37[i]) * tempf;
    params->internal_loop[i]  = internal_loopdH[i] - (internal_loopdH[i] - internal_loop37[i]) * tempf;
  }
  params->lxc = lxc37*tempf;
  for (; i<=MAXLOOP; i++) {
    params->bulge[i] = params->bulge[30]+(int)(params->lxc*log((double)(i)/30.));
    params->internal_loop[i] = params->internal_loop[30]+(int)(params->lxc*log((double)(i)/30.));
  }

  params->ninio[2] = niniodH - (niniodH - ninio37) * tempf;

  params->TripleC = TripleCdH - (TripleCdH - TripleC37) * tempf;
  params->MultipleCA = MultipleCAdH - (MultipleCAdH - MultipleCA37) * tempf;
  params->MultipleCB = MultipleCBdH - (MultipleCBdH - MultipleCB37) * tempf;

  for (i=0; (i*7)<strlen(Tetraloops); i++)
    params->Tetraloop_E[i] = TetraloopdH[i] - (TetraloopdH[i]-Tetraloop37[i])*tempf;
  for (i=0; (i*5)<strlen(Triloops); i++)
    params->Triloop_E[i] =  TriloopdH[i] - (TriloopdH[i]-Triloop37[i])*tempf;
  for (i=0; (i*9)<strlen(Hexaloops); i++)
    params->Hexaloop_E[i] =  HexaloopdH[i] - (HexaloopdH[i]-Hexaloop37[i])*tempf;

  params->TerminalAU = TerminalAUdH - (TerminalAUdH - TerminalAU37) * tempf;
 for(i=3; i<7;i++)
    params->TerminalAU_t[i] =(int) (TerminalAUdH - (TerminalAUdH - TerminalAU37) * tempf);
  params->DuplexInit = DuplexInitdH - (DuplexInitdH - DuplexInit37) *tempf;

  params->MLbase = ML_BASEdH - (ML_BASEdH - ML_BASE37) * tempf;

  for (i=0; i<=NBPAIRS; i++)
    params->MLintern[i] = ML_interndH - (ML_interndH - ML_intern37) * tempf;

  params->MLclosing = ML_closingdH - (ML_closingdH - ML_closing37) * tempf;


  /* stacks    G(T) = H - [H - G(T0)]*T/T0 */
  for (i=0; i<=NBPAIRS; i++)
    for (j=0; j<=NBPAIRS; j++)
      params->stack[i][j] = stackdH[i][j] - (stackdH[i][j] - stack37[i][j])*tempf;

  /* mismatches */
  for (i=0; i<=NBPAIRS; i++)
    for (j=0; j<5; j++)
      for (k=0; k<5; k++) {
        int mm;
        params->mismatchI[i][j][k]    = mismatchIdH[i][j][k] - (mismatchIdH[i][j][k] - mismatchI37[i][j][k])*tempf;
        params->mismatchH[i][j][k]    = mismatchHdH[i][j][k] - (mismatchHdH[i][j][k] - mismatchH37[i][j][k])*tempf;
        params->mismatch1nI[i][j][k]  = mismatch1nIdH[i][j][k]-(mismatch1nIdH[i][j][k]-mismatch1nI37[i][j][k])*tempf;/* interior nx1 loops */
        params->mismatch23I[i][j][k]  = mismatch23IdH[i][j][k]-(mismatch23IdH[i][j][k]-mismatch23I37[i][j][k])*tempf;/* interior 2x3 loops */
        if(md.dangles){
          mm                      = mismatchMdH[i][j][k] - (mismatchMdH[i][j][k] - mismatchM37[i][j][k])*tempf;
          params->mismatchM[i][j][k]    = (mm > 0) ? 0 : mm;
          mm                      = mismatchExtdH[i][j][k] - (mismatchExtdH[i][j][k] - mismatchExt37[i][j][k])*tempf;
          params->mismatchExt[i][j][k]  = (mm > 0) ? 0 : mm;
        }
        else{
          params->mismatchM[i][j][k] = params->mismatchExt[i][j][k] = 0;
        }
      }

  /* dangles */
  for (i=0; i<=NBPAIRS; i++)
    for (j=0; j<5; j++) {
      int dd;
      dd = dangle5_dH[i][j] - (dangle5_dH[i][j] - dangle5_37[i][j])*tempf;
      params->dangle5[i][j] = (dd>0) ? 0 : dd;  /* must be <= 0 */
      dd = dangle3_dH[i][j] - (dangle3_dH[i][j] - dangle3_37[i][j])*tempf;
      params->dangle3[i][j] = (dd>0) ? 0 : dd;  /* must be <= 0 */
    }
  /* interior 1x1 loops */
  for (i=0; i<=NBPAIRS; i++)
    for (j=0; j<=NBPAIRS; j++)
      for (k=0; k<5; k++)
        for (l=0; l<5; l++)
          params->int11[i][j][k][l] = int11_dH[i][j][k][l] - (int11_dH[i][j][k][l] - int11_37[i][j][k][l])*tempf;

  /* interior 2x1 loops */
  for (i=0; i<=NBPAIRS; i++)
    for (j=0; j<=NBPAIRS; j++)
      for (k=0; k<5; k++)
        for (l=0; l<5; l++) {
          int m;
          for (m=0; m<5; m++)
            params->int21[i][j][k][l][m] = int21_dH[i][j][k][l][m] - (int21_dH[i][j][k][l][m] - int21_37[i][j][k][l][m])*tempf;
        }
  /* interior 2x2 loops */
  for (i=0; i<=NBPAIRS; i++)
    for (j=0; j<=NBPAIRS; j++)
      for (k=0; k<5; k++)
        for (l=0; l<5; l++) {
          int m,n;
          for (m=0; m<5; m++)
            for (n=0; n<5; n++)
              params->int22[i][j][k][l][m][n] = int22_dH[i][j][k][l][m][n] - (int22_dH[i][j][k][l][m][n]-int22_37[i][j][k][l][m][n])*tempf;
        }

  strncpy(params->Tetraloops, Tetraloops, 281);
  strncpy(params->Triloops, Triloops, 241);
  strncpy(params->Hexaloops, Hexaloops, 361);

  params->id = ++id;
  return params;
}

PUBLIC paramT *scale_parameters_hybrid(void){
  model_detailsT  md;
  set_model_details(&md);
  return get_scaled_parameters_hybrid(temperature, md);
}

PUBLIC paramT *get_scaled_parameters_hybrid(double temperature,
                                            model_detailsT md){

  int i,j,k,l,m,n;
  double tempf;
  paramT *params;

  params = (paramT *)space(sizeof(paramT));

  /* store the model details */
  params->model_details = md;
  params->temperature   = temperature;
  tempf                 = ((temperature+K0)/Tmeasure);

  /****************************/
  /* Hairpin stuff            */
  /****************************/

  /* RNA free energies */
  for (i=0; i<31; i++)
    params->hairpin[i]  = hairpindH[i] - (hairpindH[i] - hairpin37[i])*tempf;
  /* RNA-DNA Hybrid free energies */
  for (i=0; i<31; i++)
    params->hairpin_RD[i]  = hairpindH_RD[i] - (hairpindH_RD[i] - hairpin37_RD[i])*tempf;
  /* DNA free energies */
  for (i=0; i<31; i++)
    params->hairpin_D[i]  = hairpindH_D[i] - (hairpindH_D[i] - hairpin37_D[i])*tempf;


  for (i=0; i<=MIN2(30,MAXLOOP); i++) {
    params->bulge[i]          = bulgedH[i] - (bulgedH[i] - bulge37[i]) * tempf;
    params->internal_loop[i]  = internal_loopdH[i] - (internal_loopdH[i] - internal_loop37[i]) * tempf;

    params->bulge_RD[i]          = bulgedH_RD[i] - (bulgedH_RD[i] - bulge37_RD[i]) * tempf;
    params->internal_loop_RD[i]  = internal_loopdH_RD[i] - (internal_loopdH_RD[i] - internal_loop37_RD[i]) * tempf;

    params->bulge_D[i]          = bulgedH_D[i] - (bulgedH_D[i] - bulge37_D[i]) * tempf;
    params->internal_loop_D[i]  = internal_loopdH_D[i] - (internal_loopdH_D[i] - internal_loop37_D[i]) * tempf;
  }

  params->lxc = lxc37*tempf;
  params->lxc_RD = lxc37_RD*tempf;
  params->lxc_D = lxc37_D*tempf;


  for (; i<=MAXLOOP; i++) {
    params->bulge[i] = params->bulge[30]+(int)(params->lxc*log((double)(i)/30.));
    params->internal_loop[i] = params->internal_loop[30]+(int)(params->lxc*log((double)(i)/30.));

    params->bulge_RD[i] = params->bulge_RD[30]+(int)(params->lxc_RD*log((double)(i)/30.));
    params->internal_loop_RD[i] = params->internal_loop_RD[30]+(int)(params->lxc_RD*log((double)(i)/30.));

    params->bulge_D[i] = params->bulge_D[30]+(int)(params->lxc_D*log((double)(i)/30.));
    params->internal_loop_D[i] = params->internal_loop_D[30]+(int)(params->lxc_D*log((double)(i)/30.));
  }

  params->ninio[2] = niniodH - (niniodH - ninio37) * tempf;
  params->ninio_RD[2] = niniodH_RD - (niniodH_RD - ninio37_RD) * tempf;
  params->ninio_D[2] = niniodH_D - (niniodH_D - ninio37_D) * tempf;

  params->TripleC = TripleCdH - (TripleCdH - TripleC37) * tempf;
  params->MultipleCA = MultipleCAdH - (MultipleCAdH - MultipleCA37) * tempf;
  params->MultipleCB = MultipleCBdH - (MultipleCBdH - MultipleCB37) * tempf;

  /* fill in the remaining Triple energies for hybrid and DNA */

  for (i=0; (i*7)<strlen(Tetraloops); i++)
    params->Tetraloop_E[i] = TetraloopdH[i] - (TetraloopdH[i]-Tetraloop37[i])*tempf;
  for (i=0; (i*5)<strlen(Triloops); i++)
    params->Triloop_E[i] =  TriloopdH[i] - (TriloopdH[i]-Triloop37[i])*tempf;
  for (i=0; (i*9)<strlen(Hexaloops); i++)
    params->Hexaloop_E[i] =  HexaloopdH[i] - (HexaloopdH[i]-Hexaloop37[i])*tempf;

  for (i=0; (i*7)<strlen(Tetraloops_RD); i++)
    params->Tetraloop_E_RD[i] = TetraloopdH_RD[i] - (TetraloopdH_RD[i]-Tetraloop37_RD[i])*tempf;
  for (i=0; (i*5)<strlen(Triloops_RD); i++)
    params->Triloop_E_RD[i] =  TriloopdH_RD[i] - (TriloopdH_RD[i]-Triloop37_RD[i])*tempf;
  for (i=0; (i*9)<strlen(Hexaloops_RD); i++)
    params->Hexaloop_E_RD[i] =  HexaloopdH_RD[i] - (HexaloopdH_RD[i]-Hexaloop37_RD[i])*tempf;

  for (i=0; (i*7)<strlen(Tetraloops_D); i++)
    params->Tetraloop_E_D[i] = TetraloopdH_D[i] - (TetraloopdH_D[i]-Tetraloop37_D[i])*tempf;
  for (i=0; (i*5)<strlen(Triloops_D); i++)
    params->Triloop_E_D[i] =  TriloopdH_D[i] - (TriloopdH_D[i]-Triloop37_D[i])*tempf;
  for (i=0; (i*9)<strlen(Hexaloops_D); i++)
    params->Hexaloop_E_D[i] =  HexaloopdH_D[i] - (HexaloopdH_D[i]-Hexaloop37_D[i])*tempf;

  params->TerminalAU = TerminalAUdH - (TerminalAUdH - TerminalAU37) * tempf;
  for(i=3; i<7;i++)
    params->TerminalAU_t[i] =(int) (TerminalAUdH - (TerminalAUdH - TerminalAU37) * tempf);
  for(i=11; i<15; i++)
    params->TerminalAU_t[i] = TerminalAUdH_RD - (TerminalAUdH_RD - TerminalAU37_RD) * tempf;
  for(i=19; i<23; i++)
    params->TerminalAU_t[i] = TerminalAUdH_RD - (TerminalAUdH_RD - TerminalAU37_RD) * tempf;
  for(i=27; i<31; i++) 
    params->TerminalAU_t[i] = TerminalAUdH_D - (TerminalAUdH_D - TerminalAU37_D) * tempf;

  params->DuplexInit = DuplexInitdH_RD - (DuplexInitdH_RD - DuplexInit37_RD) *tempf;

  params->MLbase = ML_BASEdH - (ML_BASEdH - ML_BASE37) * tempf;
  params->MLbase_D = ML_BASEdH_D - (ML_BASEdH_D - ML_BASE37_D) * tempf;
  for (i=0; i<=NBPAIRS; i++){

    params->MLintern[i] = ML_interndH - (ML_interndH - ML_intern37) * tempf;
  }
  for (i=NBPAIRS+1; i<=3*NBPAIRS+2; i++){
    params->MLintern[i] = ML_interndH_RD - (ML_interndH_RD - ML_intern37_RD) * tempf;
  }
  for (i=3*NBPAIRS+3; i<=4*NBPAIRS+3; i++){
    params->MLintern[i] = ML_interndH_D - (ML_interndH_D - ML_intern37_D) * tempf;
  }

  params->MLclosing = ML_closingdH - (ML_closingdH - ML_closing37) * tempf;
  params->MLclosing_RD = ML_closingdH_RD - (ML_closingdH_RD - ML_closing37_RD) * tempf;
  params->MLclosing_D = ML_closingdH_D - (ML_closingdH_D - ML_closing37_D) * tempf;


  /* stacks    G(T) = H - [H - G(T0)]*T/T0 */
  for (i=0; i<=7; i++){
    for (j=0; j<=7; j++)
      params->stack[i][j] = stackdH[i][j] - (stackdH[i][j] - stack37[i][j])*tempf; /* RNA RNA */
  }
  for (i=8; i<=15; i++){
    for (j=8; j<=15; j++)
      params->stack[i][j+8] = stackdH_RD[i-8][j-8] - (stackdH_RD[i-8][j-8] - stack37_RD[i-8][j-8])*tempf; /* RNA DNA */
  }
  for (i=16; i<=23; i++){
    for (j=16; j<=23; j++)
      params->stack[i][j-8] = stackdH_RD[j-16][i-16] - (stackdH_RD[j-16][i-16] - stack37_RD[j-16][i-16])*tempf; /* RNA DNA *//* params->stack[j-8][i-8] ;*/ /* DNA RNA */
  }
  for (i=24; i<=31; i++){
    for (j=24; j<=31; j++)
      params->stack[i][j] = stackdH_D[i-24][j-24] - (stackdH_D[i-24][j-24] - stack37_D[i-24][j-24])*tempf; /* DNA DNA */
  }


  /* mismatches */
  for (i=0; i<=NBPAIRS; i++)
    for (j=0; j<5; j++)
      for (k=0; k<5; k++) {
        int mm;
        params->mismatchI[i][j][k]    = mismatchIdH[i][j][k] - (mismatchIdH[i][j][k] - mismatchI37[i][j][k])*tempf;
        params->mismatchH[i][j][k]    = mismatchHdH[i][j][k] - (mismatchHdH[i][j][k] - mismatchH37[i][j][k])*tempf;
        params->mismatch1nI[i][j][k]  = mismatch1nIdH[i][j][k]-(mismatch1nIdH[i][j][k]-mismatch1nI37[i][j][k])*tempf;/* interior nx1 loops */
        params->mismatch23I[i][j][k]  = mismatch23IdH[i][j][k]-(mismatch23IdH[i][j][k]-mismatch23I37[i][j][k])*tempf;/* interior 2x3 loops */
        if(md.dangles){
          mm                      = mismatchMdH[i][j][k] - (mismatchMdH[i][j][k] - mismatchM37[i][j][k])*tempf;
          params->mismatchM[i][j][k]    = (mm > 0) ? 0 : mm;
          mm                      = mismatchExtdH[i][j][k] - (mismatchExtdH[i][j][k] - mismatchExt37[i][j][k])*tempf;
          params->mismatchExt[i][j][k]  = (mm > 0) ? 0 : mm;
        }
        else{
          params->mismatchM[i][j][k] = params->mismatchExt[i][j][k] = 0;
        }
      }
  for (i=8; i<=15; i++)
    for (j=0; j<5; j++)
      for (k=8; k<13; k++) {
        int mm;
        params->mismatchI[i][j][k]    = mismatchIdH_RD[i-8][j][k-8] - (mismatchIdH_RD[i-8][j][k-8] - mismatchI37_RD[i-8][j][k-8])*tempf;
        params->mismatchH[i][j][k]    = mismatchHdH_RD[i-8][j][k-8] - (mismatchHdH_RD[i-8][j][k-8] - mismatchH37_RD[i-8][j][k-8])*tempf;
        params->mismatch1nI[i][j][k]  = mismatch1nIdH_RD[i-8][j][k-8]-(mismatch1nIdH_RD[i-8][j][k-8]-mismatch1nI37_RD[i-8][j][k-8])*tempf;/* interior nx1 loops */
        params->mismatch23I[i][j][k]  = mismatch23IdH_RD[i-8][j][k-8]-(mismatch23IdH_RD[i-8][j][k-8]-mismatch23I37_RD[i-8][j][k-8])*tempf;/* interior 2x3 loops */
        if(dangles){
          mm                      = mismatchMdH_RD[i-8][j][k-8] - (mismatchMdH_RD[i-8][j][k-8] - mismatchM37_RD[i-8][j][k-8])*tempf;
          params->mismatchM[i][j][k]    = (mm > 0) ? 0 : mm;
          mm                      = mismatchExtdH_RD[i-8][j][k-8] - (mismatchExtdH_RD[i-8][j][k-8] - mismatchExt37_RD[i-8][j][k-8])*tempf;
          params->mismatchExt[i][j][k]  = (mm > 0) ? 0 : mm;
        }
        else{
          params->mismatchM[i][j][k] = params->mismatchExt[i][j][k] = 0;
        }
      }
  for (i=16; i<=23; i++)
    for (j=8; j<13; j++)
      for (k=0; k<5; k++) {
        int mm;
        params->mismatchI[i][j][k]    = mismatchIdH_RD[i-16][j-8][k] - (mismatchIdH_RD[i-16][j-8][k] - mismatchI37_RD[i-16][j-8][k])*tempf;
        params->mismatchH[i][j][k]    = mismatchHdH_RD[i-16][j-8][k] - (mismatchHdH_RD[i-16][j-8][k] - mismatchH37_RD[i-16][j-8][k])*tempf;
        params->mismatch1nI[i][j][k]  = mismatch1nIdH_RD[i-16][j-8][k]-(mismatch1nIdH_RD[i-16][j-8][k]-mismatch1nI37_RD[i-16][j-8][k])*tempf;/* interior nx1 loops */
        params->mismatch23I[i][j][k]  = mismatch23IdH_RD[i-16][j-8][k]-(mismatch23IdH_RD[i-16][j-8][k]-mismatch23I37_RD[i-16][j-8][k])*tempf;/* interior 2x3 loops */
        if(dangles){
          mm                      = mismatchMdH_RD[i-16][j-8][k] - (mismatchMdH_RD[i-16][j-8][k] - mismatchM37_RD[i-16][j-8][k])*tempf;
          params->mismatchM[i][j][k]    = (mm > 0) ? 0 : mm;
          mm                      = mismatchExtdH_RD[i-16][j-8][k] - (mismatchExtdH_RD[i-16][j-8][k] - mismatchExt37_RD[i-16][j-8][k])*tempf;
          params->mismatchExt[i][j][k]  = (mm > 0) ? 0 : mm;
        }
        else{
          params->mismatchM[i][j][k] = params->mismatchExt[i][j][k] = 0;
        }
      }
  for (i=24; i<=31; i++)
    for (j=8; j<13; j++)
      for (k=8; k<13; k++) {
        int mm;
        params->mismatchI[i][j][k]    = mismatchIdH_D[i-24][j-8][k-8] - (mismatchIdH_D[i-24][j-8][k-8] - mismatchI37_D[i-24][j-8][k-8])*tempf;
        params->mismatchH[i][j][k]    = mismatchHdH_D[i-24][j-8][k-8] - (mismatchHdH_D[i-24][j-8][k-8] - mismatchH37_D[i-24][j-8][k-8])*tempf;
        params->mismatch1nI[i][j][k]  = mismatch1nIdH_D[i-24][j-8][k-8]-(mismatch1nIdH_D[i-24][j-8][k-8]-mismatch1nI37_D[i-24][j-8][k-8])*tempf;/* interior nx1 loops */
        params->mismatch23I[i][j][k]  = mismatch23IdH_D[i-24][j-8][k-8]-(mismatch23IdH_D[i-24][j-8][k-8]-mismatch23I37_D[i-24][j-8][k-8])*tempf;/* interior 2x3 loops */
        if(dangles){
          mm                      = mismatchMdH_D[i-24][j-8][k-8] - (mismatchMdH_D[i-24][j-8][k-8] - mismatchM37_D[i-24][j-8][k-8])*tempf;
          params->mismatchM[i][j][k]    = (mm > 0) ? 0 : mm;
          mm                      = mismatchExtdH_D[i-24][j-8][k-8] - (mismatchExtdH_D[i-24][j-8][k-8] - mismatchExt37_D[i-24][j-8][k-8])*tempf;
          params->mismatchExt[i][j][k]  = (mm > 0) ? 0 : mm;
        }
        else{
          params->mismatchM[i][j][k] = params->mismatchExt[i][j][k] = 0;
        }
      }

  /* dangles */
  for (i=0; i<=NBPAIRS; i++)
    for (j=0; j<5; j++) {
      int dd = dangle5_dH[i][j] - (dangle5_dH[i][j] - dangle5_37[i][j])*tempf;
      params->dangle5[i][j] = (dd>0) ? 0 : dd;  /* must be <= 0 */
      dd = dangle3_dH[i][j] - (dangle3_dH[i][j] - dangle3_37[i][j])*tempf;
      params->dangle3[i][j] = (dd>0) ? 0 : dd;  /* must be <= 0 */
    }
  for (i=8; i<=15; i++){
    for (j=0; j<5; j++) {
      int dd = dangle5_dH[i-8][j] - (dangle5_dH[i-8][j] - dangle5_37[i-8][j])*tempf;
      params->dangle5[i][j] = (dd>0) ? 0 : dd;  /* must be <= 0 */
    }
    for(j=8; j<13;j++){
      int dd = dangle3_dH_D[i-8][j-8] - (dangle3_dH_D[i-8][j-8] - dangle3_37_D[i-8][j-8])*tempf;
      params->dangle3[i][j] = (dd>0) ? 0 : dd;  /* must be <= 0 */
    }
  }
  for (i=16; i<=23; i++){
    for (j=0; j<5; j++) {
      int dd = dangle5_dH_D[i-16][j] - (dangle5_dH_D[i-16][j] - dangle5_37_D[i-16][j])*tempf;
      params->dangle5[i][j] = (dd>0) ? 0 : dd;  /* must be <= 0 */
    }
    for(j=8; j<13;j++){
      int dd = dangle3_dH[i-16][j-8] - (dangle3_dH[i-16][j-8] - dangle3_37[i-16][j-8])*tempf;
      params->dangle3[i][j] = (dd>0) ? 0 : dd;  /* must be <= 0 */
    }
  }
  for (i=24; i<=31; i++)
    for (j=8; j<13; j++) {
      int dd = dangle5_dH_D[i-24][j-8] - (dangle5_dH_D[i-24][j-8] - dangle5_37_D[i-24][j-8])*tempf;
      params->dangle5[i][j] = (dd>0) ? 0 : dd;  /* must be <= 0 */
      dd = dangle3_dH_D[i-24][j-8] - (dangle3_dH_D[i-24][j-8] - dangle3_37_D[i-24][j-8])*tempf;
      params->dangle3[i][j] = (dd>0) ? 0 : dd;  /* must be <= 0 */
    }


  /* interior 1x1 loops */
  for (i=0; i<=NBPAIRS; i++)
    for (j=0; j<=NBPAIRS; j++)
      for (k=0; k<5; k++)
        for (l=0; l<5; l++)
          params->int11[i][j][k][l] = int11_dH[i][j][k][l] - (int11_dH[i][j][k][l] - int11_37[i][j][k][l])*tempf;

  for (i=8; i<=15; i++)
    for (j=8; j<=15; j++)
      for (k=0; k<5; k++)
        for (l=8; l<13; l++)
          params->int11[i][j+8/*?*/][k][l] = int11_dH_RD[i-8][j-8][k][l-8] - (int11_dH_RD[i-8][j-8][k][l-8] - int11_37_RD[i-8][j-8][k][l-8])*tempf;

  for (i=16; i<=23; i++)
    for (j=16; j<=23; j++)
      for (k=8; k<13; k++)
        for (l=0; l<5; l++)
          params->int11[i][j-8/*?*/][k][l] = int11_dH_RD[i-16][j-16][k-8][l] - (int11_dH_RD[i-16][j-16][k-8][l] - int11_37_RD[i-16][j-16][k-8][l])*tempf;

  for (i=24; i<=31; i++)
    for (j=24; j<=31; j++)
      for (k=8; k<13; k++)
        for (l=8; l<13; l++)
          params->int11[i][j][k][l] = int11_dH_D[i-24][j-24][k-8][l-8] - (int11_dH_D[i-24][j-24][k-8][l-8] - int11_37_D[i-24][j-24][k-8][l-8])*tempf;

  /* interior 2x1 loops */
  for (i=0; i<=NBPAIRS; i++)
    for (j=0; j<=NBPAIRS; j++)
      for (k=0; k<5; k++)
        for (l=0; l<5; l++) {
          for (m=0; m<5; m++)
            params->int21[i][j][k][l][m] = int21_dH[i][j][k][l][m] - (int21_dH[i][j][k][l][m] - int21_37[i][j][k][l][m])*tempf;
        }
  for (i=8; i<=15; i++)
    for (j=8; j<=15; j++)
      for (k=0; k<5; k++)
        for (l=8; l<13; l++) {
          for (m=8; m<13; m++)
            params->int21[i][j+8/*?*/][k][l][m] = int21_dH_RD[i-8][j-8][k][l-8][m-8] - (int21_dH_RD[i-8][j-8][k][l-8][m-8] - int21_37_RD[i-8][j-8][k][l-8][m-8])*tempf;
        }
  for (i=16; i<=23; i++)
    for (j=16; j<=23; j++)
      for (k=8; k<13; k++)
        for (l=0; l<5; l++) {
          for (m=0; m<5; m++)
            params->int21[i][j-8/*?*/][k][l][m] = int21_dH_RD[i-16][j-16][k-8][l][m] - (int21_dH_RD[i-16][j-16][k-8][l][m] - int21_37_RD[i-16][j-16][k-8][l][m])*tempf;
        }
  for (i=24; i<=31; i++)
    for (j=24; j<=31; j++)
      for (k=8; k<13; k++)
        for (l=8; l<13; l++) {
          for (m=8; m<13; m++)
            params->int21[i][j][k][l][m] = int21_dH_D[i-24][j-24][k-8][l-8][m-8] - (int21_dH_D[i-24][j-24][k-8][l-8][m-8] - int21_37_D[i-24][j-24][k-8][l-8][m-8])*tempf;
        }



  /* interior 2x2 loops */
  for (i=0; i<=NBPAIRS; i++)
    for (j=0; j<=NBPAIRS; j++)
      for (k=0; k<5; k++)
        for (l=0; l<5; l++) {
          for (m=0; m<5; m++)
            for (n=0; n<5; n++)
              params->int22[i][j][k][l][m][n] = int22_dH[i][j][k][l][m][n] - (int22_dH[i][j][k][l][m][n]-int22_37[i][j][k][l][m][n])*tempf;
        }

  for (i=8; i<=15; i++)
    for (j=8; j<=15; j++)
      for (k=0; k<5; k++)
        for (l=0; l<5; l++) {
          for (m=8; m<13; m++)
            for (n=8; n<13; n++)
              params->int22[i][j+8/*?*/][k][l][m][n] = int22_dH_RD[i-8][j-8][k][l][m-8][n-8] - (int22_dH_RD[i-8][j-8][k][l][m-8][n-8]-int22_37_RD[i-8][j-8][k][l][m-8][n-8])*tempf;
        }

  for (i=16; i<=23; i++)
    for (j=16; j<=23; j++)
      for (k=8; k<13; k++)
        for (l=8; l<13; l++) {
          for (m=0; m<5; m++)
            for (n=0; n<5; n++)
              params->int22[i][j-8/*?*/][k][l][m][n] = int22_dH_RD[i-16][j-16][k-8][l-8][m][n] - (int22_dH_RD[i-16][j-16][k-8][l-8][m][n]-int22_37_RD[i-16][j-16][k-8][l-8][m][n])*tempf;
        }
 
  for (i=24; i<=31; i++)
    for (j=24; j<=31; j++)
      for (k=8; k<13; k++)
        for (l=8; l<13; l++) {
          for (m=8; m<13; m++)
            for (n=8; n<13; n++){
              params->int22[i][j][k][l][m][n] = int22_dH_D[i-24][j-24][k-8][l-8][m-8][n-8] - (int22_dH_D[i-24][j-24][k-8][l-8][m-8][n-8] - int22_37_D[i-24][j-24][k-8][l-8][m-8][n-8])*tempf;
            }
        }

  strncpy(params->Tetraloops, Tetraloops, 281);
  strncpy(params->Triloops, Triloops, 241);
  strncpy(params->Hexaloops, Hexaloops, 361);

  strncpy(params->Tetraloops_RD, Tetraloops_RD, 281);
  strncpy(params->Triloops_RD, Triloops_RD, 241);
  strncpy(params->Hexaloops_RD, Hexaloops_RD, 361);

  strncpy(params->Tetraloops_D, Tetraloops_D, 281);
  strncpy(params->Triloops_D, Triloops_D, 241);
  strncpy(params->Hexaloops_D, Hexaloops_D, 361);

  params->id = ++id;
  return params;
}

/*------------------------------------------------------------------------*/
#define SCALE 10
/**
*** dangling ends should never be destabilizing, i.e. expdangle>=1<BR>
*** specific heat needs smooth function (2nd derivative)<BR>
*** we use a*(sin(x+b)+1)^2, with a=2/(3*sqrt(3)), b=Pi/6-sqrt(3)/2,
*** in the interval b<x<sqrt(3)/2
*/
#define SMOOTH(X) ((X)/SCALE<-1.2283697)?0:(((X)/SCALE>0.8660254)?(X):\
          SCALE*0.38490018*(sin((X)/SCALE-0.34242663)+1)*(sin((X)/SCALE-0.34242663)+1))

/* #define SMOOTH(X) ((X)<0 ? 0 : (X)) */


PUBLIC pf_paramT *get_scaled_pf_parameters(void){
  model_detailsT  md;
  set_model_details(&md);
  return get_boltzmann_factors(temperature, 1.0, md, pf_scale);
}

PUBLIC pf_paramT *get_boltzmann_factors(double temp,
                                        double betaScale,
                                        model_detailsT md,
                                        double pf_scale){

  unsigned  int i, j, k, l;
  double        kT, TT;
  double        GT;
  pf_paramT     *pf;

  pf                = (pf_paramT *)space(sizeof(pf_paramT));
  pf->model_details = md;
  pf->temperature   = temp;
  pf->alpha         = betaScale;
  pf->kT = kT       = betaScale*(temp+K0)*GASCONST;   /* kT in cal/mol  */
  pf->pf_scale      = pf_scale;
  TT                = (temp+K0)/(Tmeasure);

  for(i = VRNA_GQUAD_MIN_STACK_SIZE; i <= VRNA_GQUAD_MAX_STACK_SIZE; i++)
    for(j = 3*VRNA_GQUAD_MIN_LINKER_LENGTH; j <= 3*VRNA_GQUAD_MAX_LINKER_LENGTH; j++){
      double GQuadAlpha_T = (double)GQuadAlphadH - (double)(GQuadAlphadH - GQuadAlpha37) * TT;
      double GQuadBeta_T = (double)GQuadBetadH - (double)(GQuadBetadH - GQuadBeta37) * TT;
      GT = ((double)GQuadAlpha_T)*((double)(i-1)) + ((double)GQuadBeta_T)*log(((double)j) - 2.);
      pf->expgquad[i][j] = exp( -GT*10./kT);
    }

  /* loop energies: hairpins, bulges, interior, mulit-loops */
  for (i=0; i<31; i++){
    GT  = hairpindH[i] - (hairpindH[i] - hairpin37[i])*TT;
    pf->exphairpin[i] = exp( -GT*10./kT);
  }

  for (i=0; i<=MIN2(30, MAXLOOP); i++) {
    GT =  bulgedH[i]- (bulgedH[i] - bulge37[i])*TT;
    pf->expbulge[i] = exp( -GT*10./kT);
    GT =  internal_loopdH[i] - (internal_loopdH[i] - internal_loop37[i])*TT;
    pf->expinternal[i] = exp( -GT*10./kT);
  }
  /* special case of size 2 interior loops (single mismatch) */
  if (james_rule) pf->expinternal[2] = exp ( -80*10./kT);

  pf->lxc = lxc37*TT;

  GT =  DuplexInitdH - (DuplexInitdH - DuplexInit37)*TT;
  pf->expDuplexInit = exp( -GT*10./kT);

  for (i=31; i<=MAXLOOP; i++) {
    GT = bulge37[30]*TT + (pf->lxc*log( i/30.));
    pf->expbulge[i] = exp( -GT*10./kT);
    GT = internal_loop37[30]*TT + (pf->lxc*log( i/30.));
    pf->expinternal[i] = exp( -GT*10./kT);
  }

  GT = niniodH - (niniodH - ninio37)*TT;
  for (j=0; j<=MAXLOOP; j++)
      pf->expninio[2][j]=exp(-MIN2(MAX_NINIO,j*GT)*10./kT);

  for (i=0; (i*7)<strlen(Tetraloops); i++) {
    GT = TetraloopdH[i] - (TetraloopdH[i]-Tetraloop37[i])*TT;
    pf->exptetra[i] = exp( -GT*10./kT);
  }
  for (i=0; (i*5)<strlen(Triloops); i++) {
    GT = TriloopdH[i] - (TriloopdH[i]-Triloop37[i])*TT;
    pf->exptri[i] = exp( -GT*10./kT);
  }
  for (i=0; (i*9)<strlen(Hexaloops); i++) {
    GT = HexaloopdH[i] - (HexaloopdH[i]-Hexaloop37[i])*TT;
    pf->exphex[i] = exp( -GT*10./kT);
  }
  GT =  ML_closingdH - (ML_closingdH - ML_closing37)*TT;
  pf->expMLclosing = exp( -GT*10./kT);

  for (i=0; i<=NBPAIRS; i++) {
    GT =  ML_interndH - (ML_interndH - ML_intern37)*TT;
    /* if (i>2) GT += TerminalAU; */
    pf->expMLintern[i] = exp( -GT*10./kT);
  }
  GT = TerminalAUdH - (TerminalAUdH - TerminalAU37)*TT;
  pf->expTermAU = exp(-GT*10./kT);

  GT = ML_BASEdH - (ML_BASEdH - ML_BASE37)*TT;

  pf->expMLbase=exp(-10.*GT/kT);


  /* if dangles==0 just set their energy to 0,
     don't let dangle energies become > 0 (at large temps),
     but make sure go smoothly to 0                        */
  for (i=0; i<=NBPAIRS; i++)
    for (j=0; j<=4; j++) {
      if (md.dangles) {
        GT = dangle5_dH[i][j] - (dangle5_dH[i][j] - dangle5_37[i][j])*TT;
        pf->expdangle5[i][j] = exp(SMOOTH(-GT)*10./kT);
        GT = dangle3_dH[i][j] - (dangle3_dH[i][j] - dangle3_37[i][j])*TT;
        pf->expdangle3[i][j] =  exp(SMOOTH(-GT)*10./kT);
      } else
        pf->expdangle3[i][j] = pf->expdangle5[i][j] = 1;
    }

  /* stacking energies */
  for (i=0; i<=NBPAIRS; i++)
    for (j=0; j<=NBPAIRS; j++) {
      GT =  stackdH[i][j] - (stackdH[i][j] - stack37[i][j])*TT;
      pf->expstack[i][j] = exp( -GT*10./kT);
    }

  /* mismatch energies */
  for (i=0; i<=NBPAIRS; i++)
    for (j=0; j<5; j++)
      for (k=0; k<5; k++) {
        GT =  mismatchIdH[i][j][k] - ( mismatchIdH[i][j][k] - mismatchI37[i][j][k])*TT;
        pf->expmismatchI[i][j][k] = exp(-GT*10.0/kT);
        GT = mismatch1nIdH[i][j][k] - (mismatch1nIdH[i][j][k] - mismatch1nI37[i][j][k])*TT;
        pf->expmismatch1nI[i][j][k] = exp(-GT*10.0/kT);
        GT = mismatchHdH[i][j][k] - (mismatchHdH[i][j][k] - mismatchH37[i][j][k])*TT;
        pf->expmismatchH[i][j][k] = exp(-GT*10.0/kT);
        if (md.dangles) {
          GT = mismatchMdH[i][j][k] - (mismatchMdH[i][j][k] - mismatchM37[i][j][k])*TT;
          pf->expmismatchM[i][j][k] = exp(SMOOTH(-GT)*10.0/kT);
          GT = mismatchExtdH[i][j][k] - (mismatchExtdH[i][j][k] - mismatchExt37[i][j][k])*TT;
          pf->expmismatchExt[i][j][k] = exp(SMOOTH(-GT)*10.0/kT);
        }
        else{
          pf->expmismatchM[i][j][k] = pf->expmismatchExt[i][j][k] = 1.;
        }
        GT = mismatch23IdH[i][j][k] - (mismatch23IdH[i][j][k] - mismatch23I37[i][j][k])*TT;
        pf->expmismatch23I[i][j][k] = exp(-GT*10.0/kT);
      }

  /* interior lops of length 2 */
  for (i=0; i<=NBPAIRS; i++)
    for (j=0; j<=NBPAIRS; j++)
      for (k=0; k<5; k++)
        for (l=0; l<5; l++) {
          GT = int11_dH[i][j][k][l] -
            (int11_dH[i][j][k][l] - int11_37[i][j][k][l])*TT;
          pf->expint11[i][j][k][l] = exp(-GT*10./kT);
        }
  /* interior 2x1 loops */
  for (i=0; i<=NBPAIRS; i++)
    for (j=0; j<=NBPAIRS; j++)
      for (k=0; k<5; k++)
        for (l=0; l<5; l++) {
          int m;
          for (m=0; m<5; m++) {
            GT = int21_dH[i][j][k][l][m] -
              (int21_dH[i][j][k][l][m] - int21_37[i][j][k][l][m])*TT;
            pf->expint21[i][j][k][l][m] = exp(-GT*10./kT);
          }
        }

  /* interior 2x2 loops */
  for (i=0; i<=NBPAIRS; i++)
    for (j=0; j<=NBPAIRS; j++)
      for (k=0; k<5; k++)
        for (l=0; l<5; l++) {
          int m,n;
          for (m=0; m<5; m++)
            for (n=0; n<5; n++) {
              GT = int22_dH[i][j][k][l][m][n] -
                (int22_dH[i][j][k][l][m][n]-int22_37[i][j][k][l][m][n])*TT;
              pf->expint22[i][j][k][l][m][n] = exp(-GT*10./kT);
            }
        }

  strncpy(pf->Tetraloops, Tetraloops, 281);
  strncpy(pf->Triloops, Triloops, 241);
  strncpy(pf->Hexaloops, Hexaloops, 361);

  return pf;
}

PUBLIC pf_paramT *get_scaled_pf_parameters_hybrid(void){
  model_detailsT  md;
  set_model_details(&md);
  return get_boltzmann_factors_hybrid(temperature, 1.0, md, pf_scale);
}

PUBLIC pf_paramT *get_boltzmann_factors_hybrid( double temperature,
                                                double betaScale,
                                                model_detailsT md,
                                                double pf_scale){

  /* scale energy parameters and pre-calculate Boltzmann weights */
  unsigned int  i, j, k, l;
  double        kT, TT;
  double        GT;
  pf_paramT     *pf;

  pf                = (pf_paramT *)space(sizeof(pf_paramT));
  pf->model_details = md;
  pf->temperature   = temperature;
  pf->alpha         = betaScale;
  pf->kT = kT       = betaScale*(temperature+K0)*GASCONST;   /* kT in cal/mol  */
  pf->pf_scale      = pf_scale;
  TT                = (temperature+K0)/(Tmeasure);

  /****************************/
  /* Hairpin stuff            */
  /****************************/

  /* RNA free energies */
  for (i=0; i<31; i++) {
    GT =  hairpindH[i] - (hairpindH[i] - hairpin37[i])*TT;
    pf->exphairpin[i] = exp( -GT*10./kT);
  }

  /* RNA-DNA Hybrid free energies */
  for (i=0; i<31; i++) {
    GT =  hairpindH_RD[i] - (hairpindH_RD[i] - hairpin37_RD[i])*TT;
    pf->exphairpin_RD[i] = exp( -GT*10./kT);
  }

  /* DNA free energies */
  for (i=0; i<31; i++) {
    GT =  hairpindH_D[i] - (hairpindH_D[i] - hairpin37_D[i])*TT;
    pf->exphairpin_D[i] = exp( -GT*10./kT);
  }


  /****************************/
  /* Intloop stuff            */
  /****************************/
  for (i=0; i<=MIN2(30, MAXLOOP); i++) {
    GT =  bulgedH[i]- (bulgedH[i] - bulge37[i])*TT;
    pf->expbulge[i] = exp( -GT*10./kT);
    GT =  internal_loopdH[i] - (internal_loopdH[i] - internal_loop37[i])*TT;
    pf->expinternal[i] = exp( -GT*10./kT);

    GT =  bulgedH_RD[i]- (bulgedH_RD[i] - bulge37_RD[i])*TT;
    pf->expbulge_RD[i] = exp( -GT*10./kT);
    GT =  internal_loopdH_RD[i] - (internal_loopdH_RD[i] - internal_loop37_RD[i])*TT;
    pf->expinternal_RD[i] = exp( -GT*10./kT);

    GT =  bulgedH_D[i]- (bulgedH_D[i] - bulge37_D[i])*TT;
    pf->expbulge_D[i] = exp( -GT*10./kT);
    GT =  internal_loopdH_D[i] - (internal_loopdH_D[i] - internal_loop37_D[i])*TT;
    pf->expinternal_D[i] = exp( -GT*10./kT);
  }

  /* special case of size 2 interior loops (single mismatch) */
  if (james_rule) pf->expinternal[2] = exp ( -80*10./kT);

  pf->lxc     = lxc37*TT;
  pf->lxc_RD  = lxc37_RD*TT;
  pf->lxc_D   = lxc37_D*TT;

  GT =  DuplexInitdH_RD - (DuplexInitdH_RD - DuplexInit37_RD)*TT;
  pf->expDuplexInit = exp( -GT*10./kT);

  for (i=31; i<=MAXLOOP; i++) {
    GT = bulge37[30]*TT + (pf->lxc*log( i/30.));
    pf->expbulge[i] = exp( -GT*10./kT);
    GT = internal_loop37[30]*TT + (pf->lxc*log( i/30.));
    pf->expinternal[i] = exp( -GT*10./kT);

    GT = bulge37_RD[30]*TT + (pf->lxc_RD*log( i/30.));
    pf->expbulge_RD[i] = exp( -GT*10./kT);
    GT = internal_loop37_RD[30]*TT + (pf->lxc_RD*log( i/30.));
    pf->expinternal_RD[i] = exp( -GT*10./kT);

    GT = bulge37_D[30]*TT + (pf->lxc_D*log( i/30.));
    pf->expbulge_D[i] = exp( -GT*10./kT);
    GT = internal_loop37_D[30]*TT + (pf->lxc_D*log( i/30.));
    pf->expinternal_D[i] = exp( -GT*10./kT);
  }

  GT = niniodH - (niniodH - ninio37)*TT;
  for (j=0; j<=MAXLOOP; j++)
    pf->expninio[2][j]=exp(-MIN2(MAX_NINIO,j*GT)*10./kT);

  GT = niniodH_RD - (niniodH_RD - ninio37_RD)*TT;
  for (j=0; j<=MAXLOOP; j++)
    pf->expninio_RD[2][j]=exp(-MIN2(MAX_NINIO,j*GT)*10./kT);

  GT = niniodH_D - (niniodH_D - ninio37_D)*TT;
  for (j=0; j<=MAXLOOP; j++)
    pf->expninio_D[2][j]=exp(-MIN2(MAX_NINIO,j*GT)*10./kT);


  for (i=0; (i*7)<strlen(Tetraloops); i++) {
    GT = TetraloopdH[i] - (TetraloopdH[i]-Tetraloop37[i])*TT;
    pf->exptetra[i] = exp( -GT*10./kT);
  }

  for (i=0; (i*7)<strlen(Tetraloops_RD); i++) {
    GT = TetraloopdH_RD[i] - (TetraloopdH_RD[i]-Tetraloop37_RD[i])*TT;
    pf->exptetra_RD[i] = exp( -GT*10./kT);
  }

  for (i=0; (i*7)<strlen(Tetraloops_D); i++) {
    GT = TetraloopdH_D[i] - (TetraloopdH_D[i]-Tetraloop37_D[i])*TT;
    pf->exptetra_D[i] = exp( -GT*10./kT);
  }

  for (i=0; (i*5)<strlen(Triloops); i++) {
    GT = TriloopdH[i] - (TriloopdH[i]-Triloop37[i])*TT;
    pf->exptri[i] = exp( -GT*10./kT);
  }

  for (i=0; (i*5)<strlen(Triloops_RD); i++) {
    GT = TriloopdH_RD[i] - (TriloopdH_RD[i]-Triloop37_RD[i])*TT;
    pf->exptri_RD[i] = exp( -GT*10./kT);
  }

  for (i=0; (i*5)<strlen(Triloops_D); i++) {
    GT = TriloopdH_D[i] - (TriloopdH_D[i]-Triloop37_D[i])*TT;
    pf->exptri_D[i] = exp( -GT*10./kT);
  }

  for (i=0; (i*9)<strlen(Hexaloops); i++) {
    GT = HexaloopdH[i] - (HexaloopdH[i]-Hexaloop37[i])*TT;
    pf->exphex[i] = exp( -GT*10./kT);
  }

  for (i=0; (i*9)<strlen(Hexaloops_RD); i++) {
    GT = HexaloopdH_RD[i] - (HexaloopdH_RD[i]-Hexaloop37_RD[i])*TT;
    pf->exphex_RD[i] = exp( -GT*10./kT);
  }

  for (i=0; (i*9)<strlen(Hexaloops_D); i++) {
    GT = HexaloopdH_D[i] - (HexaloopdH_D[i]-Hexaloop37_D[i])*TT;
    pf->exphex_D[i] = exp( -GT*10./kT);
  }


  GT =  ML_closingdH - (ML_closingdH - ML_closing37)*TT;
  pf->expMLclosing = exp( -GT*10./kT);

  GT =  ML_closingdH_RD - (ML_closingdH_RD - ML_closing37_RD)*TT;
  pf->expMLclosing_RD = exp( -GT*10./kT);

  GT =  ML_closingdH_D - (ML_closingdH_D - ML_closing37_D)*TT;
  pf->expMLclosing_D = exp( -GT*10./kT);

  for (i=0; i<=NBPAIRS; i++) {
    GT =  ML_interndH - (ML_interndH - ML_intern37)*TT;
    pf->expMLintern[i] = exp( -GT*10./kT);
  }
  for (i=NBPAIRS + 1; i<=3*NBPAIRS+2; i++) {
    GT =  ML_interndH_RD - (ML_interndH_RD - ML_intern37_RD)*TT;
    pf->expMLintern[i] = exp( -GT*10./kT);
  }
  for (i=3*NBPAIRS+3; i<=4*NBPAIRS+3; i++) {
    GT =  ML_interndH_D - (ML_interndH_D - ML_intern37_D)*TT;
    pf->expMLintern[i] = exp( -GT*10./kT);
  }

  GT = TerminalAUdH - (TerminalAUdH - TerminalAU37)*TT;
  pf->expTermAU = exp(-GT*10./kT);

  for(i=0; i<32;i++) /* init */
    pf->expTermAU_t[i] = 1.0;

  for(i=3; i<7;i++){
    GT = TerminalAUdH - (TerminalAUdH - TerminalAU37)*TT;
    pf->expTermAU_t[i] = exp(-GT*10./kT);
  }
  for(i=11; i<15; i++){
    GT = TerminalAUdH_RD - (TerminalAUdH_RD - TerminalAU37_RD)*TT;
    pf->expTermAU_t[i] = exp(-GT*10./kT);
  }
  for(i=19; i<23; i++){
    GT = TerminalAUdH_RD - (TerminalAUdH_RD - TerminalAU37_RD)*TT;
    pf->expTermAU_t[i] = exp(-GT*10./kT);
  }
  for(i=27; i<31; i++){ 
    GT = TerminalAUdH_D - (TerminalAUdH_D - TerminalAU37_D)*TT;
    pf->expTermAU_t[i] = exp(-GT*10./kT);
  }

  GT = ML_BASEdH - (ML_BASEdH - ML_BASE37)*TT;
  pf->expMLbase=exp(-10.*GT/kT);

  GT = ML_BASEdH_D - (ML_BASEdH_D - ML_BASE37_D)*TT;
  pf->expMLbase_D=exp(-10.*GT/kT);


  /* if dangles==0 just set their energy to 0,
     don't let dangle energies become > 0 (at large temps),
     but make sure go smoothly to 0                        */
  for (i=0; i < 32; i++) /* init */
    for (j=0; j < 15; j++)
      pf->expdangle5[i][j] = pf->expdangle3[i][j] =  1.;

  for (i=0; i<=NBPAIRS; i++)
    for (j=0; j<5; j++) {
      if (dangles) {

        GT = dangle5_dH[i][j] - (dangle5_dH[i][j] - dangle5_37[i][j])*TT;
        pf->expdangle5[i][j] = exp(SMOOTH(-GT)*10./kT);

        GT = dangle3_dH[i][j] - (dangle3_dH[i][j] - dangle3_37[i][j])*TT;
        pf->expdangle3[i][j] =  exp(SMOOTH(-GT)*10./kT);

      } else
        pf->expdangle3[i][j] = pf->expdangle5[i][j] = 1;
    }

  for (i=8; i<=15; i++)
    for (j=0; j<5; j++) {
      if (dangles) {

        GT = dangle5_dH[i-8][j] - (dangle5_dH[i-8][j] - dangle5_37[i-8][j])*TT;
        pf->expdangle5[i][j] = exp(SMOOTH(-GT)*10./kT);
      } else
        pf->expdangle5[i][j] = 1;
    }
    for(j=8; j<13;j++){
      if (dangles) {

        GT = dangle3_dH_D[i-8][j-8] - (dangle3_dH_D[i-8][j-8] - dangle3_37_D[i-8][j-8])*TT;
        pf->expdangle3[i][j] =  exp(SMOOTH(-GT)*10./kT);

      } else
        pf->expdangle3[i][j] = 1;
    }

  for (i=16; i<=23; i++)
    for (j=0; j<=4; j++) {
      if (dangles) {

        GT = dangle5_dH_D[i-16][j] - (dangle5_dH_D[i-16][j] - dangle5_37_D[i-16][j])*TT;
        pf->expdangle5[i][j] = exp(SMOOTH(-GT)*10./kT);
      } else
        pf->expdangle5[i][j] = 1;
    }
    for(j=8; j<13;j++){
      if (dangles) {

        GT = dangle3_dH[i-16][j-8] - (dangle3_dH[i-16][j-8] - dangle3_37[i-16][j-8])*TT;
        pf->expdangle3[i][j] =  exp(SMOOTH(-GT)*10./kT);

      } else
        pf->expdangle3[i][j] = 1;
    }

  for (i=24; i<=31; i++)
    for (j=8; j<13; j++) {
      if (dangles) {

        GT = dangle5_dH_D[i-24][j-8] - (dangle5_dH_D[i-24][j-8] - dangle5_37_D[i-24][j-8])*TT;
        pf->expdangle5[i][j] = exp(SMOOTH(-GT)*10./kT);

        GT = dangle3_dH_D[i-24][j-8] - (dangle3_dH_D[i-24][j-8] - dangle3_37_D[i-24][j-8])*TT;
        pf->expdangle3[i][j] =  exp(SMOOTH(-GT)*10./kT);

      } else
        pf->expdangle3[i][j] = pf->expdangle5[i][j] = 1;
    }


  /* stacking energies */

  /* init should not be necessary
  for (i=0; i< 32; i++)
    for (j=0; j < 32; j++)
      pf->expstack[i][j] = 1.;
  */

  for (i=0; i<=NBPAIRS; i++) /* RNA RNA */
    for (j=0; j<=NBPAIRS; j++) {
      GT =  stackdH[i][j] - (stackdH[i][j] - stack37[i][j])*TT;
      pf->expstack[i][j] = exp( -GT*10./kT);
    }

  for (i=8; i<=15; i++) /* RNA DNA */
    for (j=8; j<=15; j++) {
      GT =  stackdH_RD[i-8][j-8] - (stackdH_RD[i-8][j-8] - stack37_RD[i-8][j-8])*TT;
      pf->expstack[i][j+8] = exp( -GT*10./kT);
    }

  for (i=16; i<=23; i++) /* RNA DNA *//* params->stack[j-8][i-8] ;*/ /* DNA RNA */
    for (j=16; j<=23; j++) {
      GT =  stackdH_RD[i-16][j-16] - (stackdH_RD[i-16][j-16] - stack37_RD[i-16][j-16])*TT;
      pf->expstack[i][j-8] = exp( -GT*10./kT);
    }

  for (i=24; i<=31; i++) /* DNA DNA */
    for (j=24; j<=31; j++) {
      GT =  stackdH_D[i-24][j-24] - (stackdH_D[i-24][j-24] - stack37_D[i-24][j-24])*TT;
      pf->expstack[i][j] = exp( -GT*10./kT);
    }



  /* mismatch energies */

  /* init should not be necessary */

  for (i=0; i<=NBPAIRS; i++)
    for (j=0; j<5; j++)
      for (k=0; k<5; k++) {
        GT =  mismatchIdH[i][j][k] - ( mismatchIdH[i][j][k] - mismatchI37[i][j][k])*TT;
        pf->expmismatchI[i][j][k] = exp(-GT*10.0/kT);
        GT = mismatch1nIdH[i][j][k] - (mismatch1nIdH[i][j][k] - mismatch1nI37[i][j][k])*TT;
        pf->expmismatch1nI[i][j][k] = exp(-GT*10.0/kT);
        GT = mismatchHdH[i][j][k] - (mismatchHdH[i][j][k] - mismatchH37[i][j][k])*TT;
        pf->expmismatchH[i][j][k] = exp(-GT*10.0/kT);
        if (dangles) {
          GT = mismatchMdH[i][j][k] - (mismatchMdH[i][j][k] - mismatchM37[i][j][k])*TT;
          pf->expmismatchM[i][j][k] = exp(SMOOTH(-GT)*10.0/kT);
          GT = mismatchExtdH[i][j][k] - (mismatchExtdH[i][j][k] - mismatchExt37[i][j][k])*TT;
          pf->expmismatchExt[i][j][k] = exp(SMOOTH(-GT)*10.0/kT);
        }
        else{
          pf->expmismatchM[i][j][k] = pf->expmismatchExt[i][j][k] = 1.;
        }
        GT = mismatch23IdH[i][j][k] - (mismatch23IdH[i][j][k] - mismatch23I37[i][j][k])*TT;
        pf->expmismatch23I[i][j][k] = exp(-GT*10.0/kT);
      }

  for (i=8; i<=15; i++)
    for (j=0; j<5; j++)
      for (k=8; k<13; k++) {
        GT =  mismatchIdH_RD[i-8][j][k-8] - ( mismatchIdH_RD[i-8][j][k-8] - mismatchI37_RD[i-8][j][k-8])*TT;
        pf->expmismatchI[i][j][k] = exp(-GT*10.0/kT);
        GT = mismatch1nIdH_RD[i-8][j][k-8] - (mismatch1nIdH_RD[i-8][j][k-8] - mismatch1nI37_RD[i-8][j][k-8])*TT;
        pf->expmismatch1nI[i][j][k] = exp(-GT*10.0/kT);
        GT = mismatchHdH_RD[i-8][j][k-8] - (mismatchHdH_RD[i-8][j][k-8] - mismatchH37_RD[i-8][j][k-8])*TT;
        pf->expmismatchH[i][j][k] = exp(-GT*10.0/kT);
        if (dangles) {
          GT = mismatchMdH_RD[i-8][j][k-8] - (mismatchMdH_RD[i-8][j][k-8] - mismatchM37_RD[i-8][j][k-8])*TT;
          pf->expmismatchM[i][j][k] = exp(SMOOTH(-GT)*10.0/kT);
          GT = mismatchExtdH_RD[i-8][j][k-8] - (mismatchExtdH_RD[i-8][j][k-8] - mismatchExt37_RD[i-8][j][k-8])*TT;
          pf->expmismatchExt[i][j][k] = exp(SMOOTH(-GT)*10.0/kT);
        }
        else{
          pf->expmismatchM[i][j][k] = pf->expmismatchExt[i][j][k] = 1.;
        }
        GT = mismatch23IdH_RD[i-8][j][k-8] - (mismatch23IdH_RD[i-8][j][k-8] - mismatch23I37_RD[i-8][j][k-8])*TT;
        pf->expmismatch23I[i][j][k] = exp(-GT*10.0/kT);
      }

  for (i=16; i<=23; i++)
    for (j=8; j<13; j++)
      for (k=0; k<5; k++) {
        GT =  mismatchIdH_RD[i-16][j-8][k] - ( mismatchIdH_RD[i-16][j-8][k] - mismatchI37_RD[i-16][j-8][k])*TT;
        pf->expmismatchI[i][j][k] = exp(-GT*10.0/kT);
        GT = mismatch1nIdH_RD[i-16][j-8][k] - (mismatch1nIdH_RD[i-16][j-8][k] - mismatch1nI37_RD[i-16][j-8][k])*TT;
        pf->expmismatch1nI[i][j][k] = exp(-GT*10.0/kT);
        GT = mismatchHdH_RD[i-16][j-8][k] - (mismatchHdH_RD[i-16][j-8][k] - mismatchH37_RD[i-16][j-8][k])*TT;
        pf->expmismatchH[i][j][k] = exp(-GT*10.0/kT);
        if (dangles) {
          GT = mismatchMdH_RD[i-16][j-8][k] - (mismatchMdH_RD[i-16][j-8][k] - mismatchM37_RD[i-16][j-8][k])*TT;
          pf->expmismatchM[i][j][k] = exp(SMOOTH(-GT)*10.0/kT);
          GT = mismatchExtdH_RD[i-16][j-8][k] - (mismatchExtdH_RD[i-16][j-8][k] - mismatchExt37_RD[i-16][j-8][k])*TT;
          pf->expmismatchExt[i][j][k] = exp(SMOOTH(-GT)*10.0/kT);
        }
        else{
          pf->expmismatchM[i][j][k] = pf->expmismatchExt[i][j][k] = 1.;
        }
        GT = mismatch23IdH_RD[i-16][j-8][k] - (mismatch23IdH_RD[i-16][j-8][k] - mismatch23I37_RD[i-16][j-8][k])*TT;
        pf->expmismatch23I[i][j][k] = exp(-GT*10.0/kT);
      }

  for (i=24; i<=31; i++)
    for (j=8; j<13; j++)
      for (k=8; k<13; k++) {
        GT =  mismatchIdH_D[i-24][j-8][k-8] - ( mismatchIdH_D[i-24][j-8][k-8] - mismatchI37_D[i-24][j-8][k-8])*TT;
        pf->expmismatchI[i][j][k] = exp(-GT*10.0/kT);
        GT = mismatch1nIdH_D[i-24][j-8][k-8] - (mismatch1nIdH_D[i-24][j-8][k-8] - mismatch1nI37_D[i-24][j-8][k-8])*TT;
        pf->expmismatch1nI[i][j][k] = exp(-GT*10.0/kT);
        GT = mismatchHdH_D[i-24][j-8][k-8] - (mismatchHdH_D[i-24][j-8][k-8] - mismatchH37_D[i-24][j-8][k-8])*TT;
        pf->expmismatchH[i][j][k] = exp(-GT*10.0/kT);
        if (dangles) {
          GT = mismatchMdH_D[i-24][j-8][k-8] - (mismatchMdH_D[i-24][j-8][k-8] - mismatchM37_D[i-24][j-8][k-8])*TT;
          pf->expmismatchM[i][j][k] = exp(SMOOTH(-GT)*10.0/kT);
          GT = mismatchExtdH_D[i-24][j-8][k-8] - (mismatchExtdH_D[i-24][j-8][k-8] - mismatchExt37_D[i-24][j-8][k-8])*TT;
          pf->expmismatchExt[i][j][k] = exp(SMOOTH(-GT)*10.0/kT);
        }
        else{
          pf->expmismatchM[i][j][k] = pf->expmismatchExt[i][j][k] = 1.;
        }
        GT = mismatch23IdH_D[i-24][j-8][k-8] - (mismatch23IdH_D[i-24][j-8][k-8] - mismatch23I37_D[i-24][j-8][k-8])*TT;
        pf->expmismatch23I[i][j][k] = exp(-GT*10.0/kT);
      }


  /* interior lops of length 2 */
  /* init should not be necessary */
  for (i=0; i<=NBPAIRS; i++)
    for (j=0; j<=NBPAIRS; j++)
      for (k=0; k<5; k++)
        for (l=0; l<5; l++) {
          GT = int11_dH[i][j][k][l] -
            (int11_dH[i][j][k][l] - int11_37[i][j][k][l])*TT;
          pf->expint11[i][j][k][l] = exp(-GT*10./kT);
        }

  for (i=8; i<=15; i++)
    for (j=8; j<=15; j++)
      for (k=0; k<5; k++)
        for (l=8; l<13; l++) {
          GT = int11_dH_RD[i-8][j-8][k][l-8] -
            (int11_dH_RD[i-8][j-8][k][l-8] - int11_37_RD[i-8][j-8][k][l-8])*TT;
          pf->expint11[i][j+8][k][l] = exp(-GT*10./kT);
        }

  for (i=16; i<=23; i++)
    for (j=16; j<=23; j++)
      for (k=8; k<13; k++)
        for (l=0; l<5; l++) {
          GT = int11_dH_RD[i-16][j-16][k-8][l] -
            (int11_dH_RD[i-16][j-16][k-8][l] - int11_37_RD[i-16][j-16][k-8][l])*TT;
          pf->expint11[i][j-8][k][l] = exp(-GT*10./kT);
        }

  for (i=24; i<=31; i++)
    for (j=24; j<=31; j++)
      for (k=8; k<13; k++)
        for (l=8; l<13; l++) {
          GT = int11_dH_D[i-24][j-24][k-8][l-8] -
            (int11_dH_D[i-24][j-24][k-8][l-8] - int11_37_D[i-24][j-24][k-8][l-8])*TT;
          pf->expint11[i][j][k][l] = exp(-GT*10./kT);
        }



  /* interior 2x1 loops */
  /* init should not be necessary */
  for (i=0; i<=NBPAIRS; i++)
    for (j=0; j<=NBPAIRS; j++)
      for (k=0; k<5; k++)
        for (l=0; l<5; l++) {
          int m;
          for (m=0; m<5; m++) {
            GT = int21_dH[i][j][k][l][m] -
              (int21_dH[i][j][k][l][m] - int21_37[i][j][k][l][m])*TT;
            pf->expint21[i][j][k][l][m] = exp(-GT*10./kT);
          }
        }

  for (i=8; i<=15; i++)
    for (j=8; j<=15; j++)
      for (k=0; k<5; k++)
        for (l=8; l<13; l++) {
          int m;
          for (m=8; m<13; m++) {
            GT = int21_dH_RD[i-8][j-8][k][l-8][m-8] -
              (int21_dH_RD[i-8][j-8][k][l-8][m-8] - int21_37_RD[i-8][j-8][k][l-8][m-8])*TT;
            pf->expint21[i][j+8][k][l][m] = exp(-GT*10./kT);
          }
        }

  for (i=16; i<=23; i++)
    for (j=16; j<=23; j++)
      for (k=8; k<13; k++)
        for (l=0; l<5; l++) {
          int m;
          for (m=0; m<5; m++) {
            GT = int21_dH_RD[i-16][j-16][k-8][l][m] -
              (int21_dH_RD[i-16][j-16][k-8][l][m] - int21_37_RD[i-16][j-16][k-8][l][m])*TT;
            pf->expint21[i][j-8][k][l][m] = exp(-GT*10./kT);
          }
        }

  for (i=24; i<=31; i++)
    for (j=24; j<=31; j++)
      for (k=8; k<13; k++)
        for (l=8; l<13; l++) {
          int m;
          for (m=8; m<13; m++) {
            GT = int21_dH_D[i-24][j-24][k-8][l-8][m-8] -
              (int21_dH_D[i-24][j-24][k-8][l-8][m-8] - int21_37_D[i-24][j-24][k-8][l-8][m-8])*TT;
            pf->expint21[i][j][k][l][m] = exp(-GT*10./kT);
          }
        }

  /* interior 2x2 loops */
  /* init should not be necessary */
  for (i=0; i<=NBPAIRS; i++)
    for (j=0; j<=NBPAIRS; j++)
      for (k=0; k<5; k++)
        for (l=0; l<5; l++) {
          int m,n;
          for (m=0; m<5; m++)
            for (n=0; n<5; n++) {
              GT = int22_dH[i][j][k][l][m][n] -
                (int22_dH[i][j][k][l][m][n]-int22_37[i][j][k][l][m][n])*TT;
              pf->expint22[i][j][k][l][m][n] = exp(-GT*10./kT);
            }
        }

  for (i=8; i<=15; i++)
    for (j=8; j<=15; j++)
      for (k=0; k<5; k++)
        for (l=0; l<5; l++) {
          int m,n;
          for (m=8; m<13; m++)
            for (n=8; n<13; n++) {
              GT = int22_dH_RD[i-8][j-8][k][l][m-8][n-8] -
                (int22_dH_RD[i-8][j-8][k][l][m-8][n-8]-int22_37_RD[i-8][j-8][k][l][m-8][n-8])*TT;
              pf->expint22[i][j+8][k][l][m][n] = exp(-GT*10./kT);
            }
        }

  for (i=16; i<=23; i++)
    for (j=16; j<=23; j++)
      for (k=8; k<13; k++)
        for (l=8; l<13; l++) {
          int m,n;
          for (m=0; m<5; m++)
            for (n=0; n<5; n++) {
              GT = int22_dH_RD[i-16][j-16][k-8][l-8][m][n] -
                (int22_dH_RD[i-16][j-16][k-8][l-8][m][n]-int22_37_RD[i-16][j-16][k-8][l-8][m][n])*TT;
              pf->expint22[i][j-8][k][l][m][n] = exp(-GT*10./kT);
            }
        }

  for (i=24; i<=31; i++)
    for (j=24; j<=31; j++)
      for (k=8; k<13; k++)
        for (l=8; l<13; l++) {
          int m,n;
          for (m=8; m<13; m++)
            for (n=8; n<13; n++) {
              GT = int22_dH_D[i-24][j-24][k-8][l-8][m-8][n-8] -
                (int22_dH_D[i-24][j-24][k-8][l-8][m-8][n-8]-int22_37_D[i-24][j-24][k-8][l-8][m-8][n-8])*TT;
              pf->expint22[i][j][k][l][m][n] = exp(-GT*10./kT);
            }
        }

  strncpy(pf->Tetraloops, Tetraloops, 281);
  strncpy(pf->Triloops, Triloops, 241);
  strncpy(pf->Hexaloops, Hexaloops, 361);

  strncpy(pf->Tetraloops_RD, Tetraloops_RD, 281);
  strncpy(pf->Triloops_RD, Triloops_RD, 241);
  strncpy(pf->Hexaloops_RD, Hexaloops_RD, 361);

  strncpy(pf->Tetraloops_D, Tetraloops_D, 281);
  strncpy(pf->Triloops_D, Triloops_D, 241);
  strncpy(pf->Hexaloops_D, Hexaloops_D, 361);

  return pf;
}

PUBLIC pf_paramT *get_scaled_alipf_parameters(unsigned int n_seq){
  model_detailsT  md;
  set_model_details(&md);
  return get_boltzmann_factors_ali(n_seq, temperature, 1.0, md, pf_scale);
}

PUBLIC pf_paramT *get_boltzmann_factors_ali(unsigned int n_seq,
                                            double temperature,
                                            double betaScale,
                                            model_detailsT md,
                                            double pf_scale){

  /* scale energy parameters and pre-calculate Boltzmann weights */
  unsigned int  i, j, k, l;
  double        kTn, TT;
  double        GT;
  pf_paramT     *pf;

  pf                = (pf_paramT *)space(sizeof(pf_paramT));
  pf->model_details = md;
  pf->alpha         = betaScale;
  pf->temperature   = temperature;
  pf->pf_scale      = pf_scale;
  pf->kT = kTn      = ((double)n_seq)*betaScale*(temperature+K0)*GASCONST;   /* kT in cal/mol  */
  TT                = (temperature+K0)/(Tmeasure);


   /* loop energies: hairpins, bulges, interior, mulit-loops */
  for (i=0; i<31; i++) {
    GT =  hairpindH[i] - (hairpindH[i] - hairpin37[i])*TT;
    pf->exphairpin[i] = exp( -GT*10./kTn);
  }
  /*add penalty for too short hairpins*/
  for (i=0; i<3; i++) {
    GT= 600/*Penalty*/*TT;
    pf->exphairpin[i] = exp( -GT*10./kTn);
  }

  for (i=0; i<=MIN2(30, MAXLOOP); i++) {
    GT =  bulgedH[i]- (bulgedH[i] - bulge37[i])*TT;
    pf->expbulge[i] = exp( -GT*10./kTn);
    GT =  internal_loopdH[i] - (internal_loopdH[i] - internal_loop37[i])*TT;
    pf->expinternal[i] = exp( -GT*10./kTn);
  }
  /* special case of size 2 interior loops (single mismatch) */
  if (james_rule) pf->expinternal[2] = exp ( -80*10./kTn);

  pf->lxc = lxc37*TT;

  GT =  DuplexInitdH - (DuplexInitdH - DuplexInit37)*TT;
  pf->expDuplexInit = exp( -GT*10./kTn);

  for (i=31; i<=MAXLOOP; i++) {
    GT = bulge37[30]*TT + (pf->lxc*log( i/30.));
    pf->expbulge[i] = exp( -GT*10./kTn);
    GT = internal_loop37[30]*TT + (pf->lxc*log( i/30.));
    pf->expinternal[i] = exp( -GT*10./kTn);
  }

  GT = niniodH - (niniodH - ninio37)*TT;
  for (j=0; j<=MAXLOOP; j++)
    pf->expninio[2][j]=exp(-MIN2(MAX_NINIO,j*GT)*10./kTn);

  for (i=0; (i*7)<strlen(Tetraloops); i++) {
    GT = TetraloopdH[i] - (TetraloopdH[i]-Tetraloop37[i])*TT;
    pf->exptetra[i] = exp( -GT*10./kTn);
  }
  for (i=0; (i*5)<strlen(Triloops); i++) {
    GT = TriloopdH[i] - (TriloopdH[i]-Triloop37[i])*TT;
    pf->exptri[i] = exp( -GT*10./kTn);
  }
  for (i=0; (i*9)<strlen(Hexaloops); i++) {
    GT = HexaloopdH[i] - (HexaloopdH[i]-Hexaloop37[i])*TT;
    pf->exphex[i] = exp( -GT*10./kTn);
  }
  GT =  ML_closingdH - (ML_closingdH - ML_closing37)*TT;
  pf->expMLclosing = exp( -GT*10./kTn);

  for (i=0; i<=NBPAIRS; i++) { /* includes AU penalty */
    GT =  ML_interndH - (ML_interndH - ML_intern37)*TT;
    /* if (i>2) GT += TerminalAU; */
    pf->expMLintern[i] = exp( -GT*10./kTn);
  }
  GT = TerminalAUdH - (TerminalAUdH - TerminalAU37)*TT;
  pf->expTermAU = exp(-GT*10./kTn);

  GT = ML_BASEdH - (ML_BASEdH - ML_BASE37)*TT;
  pf->expMLbase=exp(-10.*GT/(kTn/n_seq));


  /* if dangle_model==0 just set their energy to 0,
     don't let dangle energies become > 0 (at large temps),
     but make sure go smoothly to 0                        */
  for (i=0; i<=NBPAIRS; i++)
    for (j=0; j<=4; j++) {
      if (md.dangles) {
        GT = dangle5_dH[i][j] - (dangle5_dH[i][j] - dangle5_37[i][j])*TT;
        pf->expdangle5[i][j] = exp(SMOOTH(-GT)*10./kTn);
        GT = dangle3_dH[i][j] - (dangle3_dH[i][j] - dangle3_37[i][j])*TT;
        pf->expdangle3[i][j] =  exp(SMOOTH(-GT)*10./kTn);
      } else
        pf->expdangle3[i][j] = pf->expdangle5[i][j] = 1;
    }

  /* stacking energies */
  for (i=0; i<=NBPAIRS; i++)
    for (j=0; j<=NBPAIRS; j++) {
      GT =  stackdH[i][j] - (stackdH[i][j] - stack37[i][j])*TT;
      pf->expstack[i][j] = exp( -GT*10./kTn);
    }

  /* mismatch energies */
  for (i=0; i<=NBPAIRS; i++)
    for (j=0; j<5; j++)
      for (k=0; k<5; k++) {
        GT =  mismatchIdH[i][j][k] - ( mismatchIdH[i][j][k] - mismatchI37[i][j][k])*TT;
        pf->expmismatchI[i][j][k] = exp(-GT*10.0/kTn);
        GT = mismatch1nIdH[i][j][k] - (mismatch1nIdH[i][j][k] - mismatch1nI37[i][j][k])*TT;
        pf->expmismatch1nI[i][j][k] = exp(-GT*10.0/kTn);
        GT = mismatchHdH[i][j][k] - (mismatchHdH[i][j][k] - mismatchH37[i][j][k])*TT;
        pf->expmismatchH[i][j][k] = exp(-GT*10.0/kTn);
        if (md.dangles) {
          GT = mismatchMdH[i][j][k] - (mismatchMdH[i][j][k] - mismatchM37[i][j][k])*TT;
          pf->expmismatchM[i][j][k] = exp(SMOOTH(-GT)*10.0/kTn);
          GT = mismatchExtdH[i][j][k] - (mismatchExtdH[i][j][k] - mismatchExt37[i][j][k])*TT;
          pf->expmismatchExt[i][j][k] = exp(SMOOTH(-GT)*10.0/kTn);
        }
        else{
          pf->expmismatchM[i][j][k] = pf->expmismatchExt[i][j][k] = 1.;
        }
        GT = mismatch23IdH[i][j][k] - (mismatch23IdH[i][j][k] - mismatch23I37[i][j][k])*TT;
        pf->expmismatch23I[i][j][k] = exp(-GT*10.0/kTn);
      }


  /* interior lops of length 2 */
  for (i=0; i<=NBPAIRS; i++)
    for (j=0; j<=NBPAIRS; j++)
      for (k=0; k<5; k++)
        for (l=0; l<5; l++) {
          GT = int11_dH[i][j][k][l] -
            (int11_dH[i][j][k][l] - int11_37[i][j][k][l])*TT;
          pf->expint11[i][j][k][l] = exp(-GT*10./kTn);
        }
  /* interior 2x1 loops */
  for (i=0; i<=NBPAIRS; i++)
    for (j=0; j<=NBPAIRS; j++)
      for (k=0; k<5; k++)
        for (l=0; l<5; l++) {
          int m;
          for (m=0; m<5; m++) {
            GT = int21_dH[i][j][k][l][m] -
              (int21_dH[i][j][k][l][m] - int21_37[i][j][k][l][m])*TT;
            pf->expint21[i][j][k][l][m] = exp(-GT*10./kTn);
          }
        }

  /* interior 2x2 loops */
  for (i=0; i<=NBPAIRS; i++)
    for (j=0; j<=NBPAIRS; j++)
      for (k=0; k<5; k++)
        for (l=0; l<5; l++) {
          int m,n;
          for (m=0; m<5; m++)
            for (n=0; n<5; n++) {
              GT = int22_dH[i][j][k][l][m][n] -
                (int22_dH[i][j][k][l][m][n]-int22_37[i][j][k][l][m][n])*TT;
              pf->expint22[i][j][k][l][m][n] = exp(-GT*10./kTn);
            }
        }

  strncpy(pf->Tetraloops, Tetraloops, 281);
  strncpy(pf->Triloops, Triloops, 241);
  strncpy(pf->Hexaloops, Hexaloops, 361);

  return pf;
}

PUBLIC pf_paramT *get_boltzmann_factor_copy(pf_paramT *par){
  pf_paramT *copy = NULL;
  if(par){
    copy = (pf_paramT *) space(sizeof(pf_paramT));
    memcpy(copy, par, sizeof(pf_paramT));
  }
  return copy;
}

PUBLIC paramT *get_parameter_copy(paramT *par){
  paramT *copy = NULL;
  if(par){
    copy = (paramT *) space(sizeof(paramT));
    memcpy(copy, par, sizeof(paramT));
  }
  return copy;
}

/*###########################################*/
/*# deprecated functions below              #*/
/*###########################################*/

PUBLIC paramT *copy_parameters(void){
  paramT *copy;
  if (p.id != id) return scale_parameters();
  else{
    copy = (paramT *) space(sizeof(paramT));
    memcpy(copy, &p, sizeof(paramT));
  }
  return copy;
}

PUBLIC paramT *set_parameters(paramT *dest){
  memcpy(&p, dest, sizeof(paramT));
  return &p;
}

PUBLIC pf_paramT *copy_pf_param(void){
  pf_paramT *copy, *new;
  if (pf.id != pf_id) return get_scaled_pf_parameters();
  else{
    copy = (pf_paramT *) space(sizeof(pf_paramT));
    memcpy(copy, &pf, sizeof(pf_paramT));
  }
  return copy;
}

PUBLIC pf_paramT *set_pf_param(paramT *dest){
  memcpy(&pf, dest, sizeof(pf_paramT));
  return &pf;
}

PUBLIC pf_paramT *scale_pf_parameters(void){
  return get_scaled_pf_parameters();
}
